-module(dig).
-compile([no_auto_import, nowarn_unused_vars, nowarn_unused_function, nowarn_nomatch]).

-export([get_path/1, parse_path_seg/1, compose/2, dig/1, dig_path_seg/1]).
-export_type([dig_decoder/0, dig_error/0, path_seg/0, path_seg_parse_error/0]).

-type dig_decoder() :: {dig_object,
        list(binary()),
        fun((gleam@dynamic:dynamic_()) -> {ok, gleam@dynamic:dynamic_()} |
            {error, list(gleam@dynamic:decode_error())})} |
    {dig_list,
        list(binary()),
        fun((gleam@dynamic:dynamic_()) -> {ok, list(gleam@dynamic:dynamic_())} |
            {error, list(gleam@dynamic:decode_error())})}.

-type dig_error() :: empty_path | {parse_path, path_seg_parse_error()}.

-type path_seg() :: {object, binary()} |
    {list, gleam@option:option(binary()), gleam@option:option(integer())}.

-type path_seg_parse_error() :: {invalid_path_seg, binary()}.

-spec get_path(dig_decoder()) -> list(binary()).
get_path(Dig_decoder) ->
    case Dig_decoder of
        {dig_object, Path, _} ->
            Path;

        {dig_list, Path@1, _} ->
            Path@1
    end.

-spec parse_object_path_seg(binary()) -> {ok, path_seg()} |
    {error, path_seg_parse_error()}.
parse_object_path_seg(Path_seg) ->
    _assert_subject = gleam@regex:from_string(<<"^\\w+$"/utf8>>),
    {ok, Obj_regex} = case _assert_subject of
        {ok, _} -> _assert_subject;
        _assert_fail ->
            erlang:error(#{gleam_error => let_assert,
                        message => <<"Assertion pattern match failed"/utf8>>,
                        value => _assert_fail,
                        module => <<"dig"/utf8>>,
                        function => <<"parse_object_path_seg"/utf8>>,
                        line => 194})
    end,
    case gleam@regex:scan(Obj_regex, Path_seg) of
        [] ->
            {error, {invalid_path_seg, Path_seg}};

        [First | _] ->
            {ok, {object, erlang:element(2, First)}}
    end.

-spec parse_list_path_seg(binary()) -> {ok, path_seg()} |
    {error, path_seg_parse_error()}.
parse_list_path_seg(Path_seg) ->
    _assert_subject = gleam@regex:from_string(<<"^(\\w*)\\[(\\d*)\\]$"/utf8>>),
    {ok, List_regex} = case _assert_subject of
        {ok, _} -> _assert_subject;
        _assert_fail ->
            erlang:error(#{gleam_error => let_assert,
                        message => <<"Assertion pattern match failed"/utf8>>,
                        value => _assert_fail,
                        module => <<"dig"/utf8>>,
                        function => <<"parse_list_path_seg"/utf8>>,
                        line => 203})
    end,
    case gleam@regex:scan(List_regex, Path_seg) of
        [] ->
            {error, {invalid_path_seg, Path_seg}};

        [First | _] ->
            gleam@result:'try'(
                begin
                    _pipe = erlang:element(3, First),
                    _pipe@1 = gleam@list:first(_pipe),
                    gleam@result:replace_error(
                        _pipe@1,
                        {invalid_path_seg, Path_seg}
                    )
                end,
                fun(Key) ->
                    gleam@result:'try'(
                        begin
                            _pipe@2 = erlang:element(3, First),
                            _pipe@3 = gleam@list:at(_pipe@2, 1),
                            gleam@result:replace_error(
                                _pipe@3,
                                {invalid_path_seg, Path_seg}
                            )
                        end,
                        fun(Index_option) -> case Index_option of
                                none ->
                                    {ok, {list, Key, none}};

                                {some, V} ->
                                    gleam@result:'try'(
                                        begin
                                            _pipe@4 = gleam@int:parse(V),
                                            gleam@result:replace_error(
                                                _pipe@4,
                                                {invalid_path_seg, Path_seg}
                                            )
                                        end,
                                        fun(Index) ->
                                            {ok, {list, Key, {some, Index}}}
                                        end
                                    )
                            end end
                    )
                end
            )
    end.

-spec parse_path_seg(binary()) -> {ok, path_seg()} |
    {error, path_seg_parse_error()}.
parse_path_seg(Path_seg) ->
    _pipe = parse_object_path_seg(Path_seg),
    gleam@result:lazy_or(_pipe, fun() -> parse_list_path_seg(Path_seg) end).

-spec map_errors(
    {ok, FMF} | {error, list(gleam@dynamic:decode_error())},
    fun((gleam@dynamic:decode_error()) -> gleam@dynamic:decode_error())
) -> {ok, FMF} | {error, list(gleam@dynamic:decode_error())}.
map_errors(Result, F) ->
    gleam@result:map_error(
        Result,
        fun(_capture) -> gleam@list:map(_capture, F) end
    ).

-spec compose(dig_decoder(), dig_decoder()) -> dig_decoder().
compose(A, B) ->
    case {A, B} of
        {{dig_object, _, A_decoder}, {dig_object, B_path, B_decoder}} ->
            {dig_object,
                B_path,
                fun(D) ->
                    gleam@result:'try'(
                        A_decoder(D),
                        fun(Dd) -> _pipe = B_decoder(Dd),
                            map_errors(
                                _pipe,
                                fun(E) -> erlang:setelement(4, E, B_path) end
                            ) end
                    )
                end};

        {{dig_object, _, A_decoder@1}, {dig_list, B_path@1, B_decoder@1}} ->
            {dig_list,
                B_path@1,
                fun(D@1) ->
                    gleam@result:'try'(
                        A_decoder@1(D@1),
                        fun(Dd@1) -> _pipe@1 = B_decoder@1(Dd@1),
                            map_errors(
                                _pipe@1,
                                fun(E@1) ->
                                    erlang:setelement(4, E@1, B_path@1)
                                end
                            ) end
                    )
                end};

        {{dig_list, _, A_decoder@2}, {dig_object, B_path@2, B_decoder@2}} ->
            {dig_list,
                B_path@2,
                fun(D@2) ->
                    gleam@result:'try'(
                        A_decoder@2(D@2),
                        fun(Dd@2) -> _pipe@2 = gleam@iterator:from_list(Dd@2),
                            _pipe@3 = gleam@iterator:map(_pipe@2, B_decoder@2),
                            _pipe@4 = gleam@iterator:map(
                                _pipe@3,
                                fun(_capture) ->
                                    map_errors(
                                        _capture,
                                        fun(E@2) ->
                                            erlang:setelement(4, E@2, B_path@2)
                                        end
                                    )
                                end
                            ),
                            _pipe@5 = gleam@iterator:to_list(_pipe@4),
                            gleam@result:all(_pipe@5) end
                    )
                end};

        {{dig_list, _, A_decoder@3}, {dig_list, B_path@3, B_decoder@3}} ->
            {dig_list,
                B_path@3,
                fun(D@3) ->
                    gleam@result:'try'(
                        A_decoder@3(D@3),
                        fun(Dd@3) ->
                            _pipe@6 = (gleam@list:flat_map(
                                Dd@3,
                                fun(Ddd) -> case B_decoder@3(Ddd) of
                                        {ok, V} ->
                                            gleam@list:map(
                                                V,
                                                fun(Field@0) -> {ok, Field@0} end
                                            );

                                        {error, E@3} ->
                                            [map_errors(
                                                    {error, E@3},
                                                    fun(E@4) ->
                                                        erlang:setelement(
                                                            4,
                                                            E@4,
                                                            B_path@3
                                                        )
                                                    end
                                                )]
                                    end end
                            )),
                            gleam@result:all(_pipe@6)
                        end
                    )
                end}
    end.

-spec dig_path_seg_with_path(list(binary()), binary()) -> {ok, dig_decoder()} |
    {error, dig_error()}.
dig_path_seg_with_path(Path, Path_seg) ->
    gleam@result:'try'(
        begin
            _pipe = parse_path_seg(Path_seg),
            gleam@result:map_error(_pipe, fun(E) -> {parse_path, E} end)
        end,
        fun(Parsed_path_seg) ->
            Path@1 = gleam@list:append(Path, [Path_seg]),
            _pipe@7 = case Parsed_path_seg of
                {object, Key} ->
                    {dig_object,
                        Path@1,
                        gleam@dynamic:field(Key, fun gleam@dynamic:dynamic/1)};

                {list, {some, Key@1}, none} ->
                    {dig_list,
                        Path@1,
                        gleam@dynamic:field(
                            Key@1,
                            gleam@dynamic:list(fun gleam@dynamic:dynamic/1)
                        )};

                {list, none, none} ->
                    {dig_list,
                        Path@1,
                        gleam@dynamic:list(fun gleam@dynamic:dynamic/1)};

                {list, {some, Key@2}, {some, Index}} ->
                    {dig_object,
                        Path@1,
                        gleam@dynamic:field(
                            Key@2,
                            fun(D) ->
                                gleam@result:'try'(
                                    begin
                                        _pipe@1 = gleam@dynamic:shallow_list(D),
                                        map_errors(
                                            _pipe@1,
                                            fun(E@1) ->
                                                erlang:setelement(
                                                    4,
                                                    E@1,
                                                    Path@1
                                                )
                                            end
                                        )
                                    end,
                                    fun(Shadow_list) -> _pipe@2 = Shadow_list,
                                        _pipe@3 = gleam@list:at(_pipe@2, Index),
                                        gleam@result:replace_error(
                                            _pipe@3,
                                            [{decode_error,
                                                    <<"index: "/utf8,
                                                        (gleam@int:to_string(
                                                            Index
                                                        ))/binary>>,
                                                    <<"missing"/utf8>>,
                                                    Path@1}]
                                        ) end
                                )
                            end
                        )};

                {list, none, {some, Index@1}} ->
                    {dig_object,
                        Path@1,
                        fun(D@1) ->
                            gleam@result:'try'(
                                begin
                                    _pipe@4 = gleam@dynamic:shallow_list(D@1),
                                    map_errors(
                                        _pipe@4,
                                        fun(E@2) ->
                                            erlang:setelement(4, E@2, Path@1)
                                        end
                                    )
                                end,
                                fun(Shallow_list) -> _pipe@5 = Shallow_list,
                                    _pipe@6 = gleam@list:at(_pipe@5, Index@1),
                                    gleam@result:replace_error(
                                        _pipe@6,
                                        [{decode_error,
                                                <<"index: "/utf8,
                                                    (gleam@int:to_string(
                                                        Index@1
                                                    ))/binary>>,
                                                <<"missing"/utf8>>,
                                                Path@1}]
                                    ) end
                            )
                        end}
            end,
            {ok, _pipe@7}
        end
    ).

-spec dig(list(binary())) -> {ok, dig_decoder()} | {error, dig_error()}.
dig(Path) ->
    case Path of
        [] ->
            {error, empty_path};

        [First | Rest] ->
            gleam@result:'try'(
                dig_path_seg_with_path([], First),
                fun(First_dig_decoder) -> _pipe = Rest,
                    gleam@list:try_fold(
                        _pipe,
                        First_dig_decoder,
                        fun(Acc_dig_decoder, It) ->
                            gleam@result:'try'(
                                dig_path_seg_with_path(
                                    get_path(Acc_dig_decoder),
                                    It
                                ),
                                fun(It_dig_decoder) ->
                                    _pipe@1 = Acc_dig_decoder,
                                    _pipe@2 = compose(_pipe@1, It_dig_decoder),
                                    {ok, _pipe@2}
                                end
                            )
                        end
                    ) end
            )
    end.

-spec dig_path_seg(binary()) -> {ok, dig_decoder()} | {error, dig_error()}.
dig_path_seg(Path_seg) ->
    dig_path_seg_with_path([], Path_seg).
